
    #include <math.h>
    #include <stdio.h>
    #include <string.h>
    #include <stdlib.h>
    #include <sys/time.h>

    #include <hmpi.h>

    #include "mxm_i.c"
    #include "mxm_benchmark_dgemm.c"
    #include "Execute_scalapack_outer_product.c"

    #include "pdgemm.c"

    int main(int argc, char **argv)
    {
        int    COORDN = 2;
        int    i, j, k, rc;
        int    *w, *h, *trow, *tcol, colw,
               myi, myj;
        double *gspeeds = NULL;

        rc = HMPI_Init(&argc, &argv);

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems initializing MPI...Exiting...\n");
        }

        int gme              = HMPI_Group_rank(HMPI_COMM_WORLD_GROUP);
        int num_of_computers = HMPI_Get_number_of_computers();
        int p                = sqrt(num_of_computers);
        int q                = num_of_computers/p;
        int reconm           = n/p;
        int reconn           = n/q;
        int min_m_n          = (reconm < reconn) ? reconm : reconn;
        int reconk;

        for (reconk = 1; reconk <= min_m_n; reconk++)
        {
           if (((reconm%reconk) != 0) || ((reconn%reconk) != 0))
           {
              continue;
           }

           /*
           * The initial benchmark
           */
           int benchmark_m = reconm;
           int benchmark_n = reconn;
           int kb          = reconk;
           int MSIZE       = n/kb;

           { 
              int output_p;
              int input_p[5] = 
                  {
                   0,
                   0,
                   benchmark_m, 
                   benchmark_n, 
                   kb
              };

              rc = HMPI_Recon_t(
                       &Benchmark_function,
                       input_p,
                       5,
                       &output_p
              ); 
   
              if  (rc != HMPI_OK)
              {
                printf("Panic: HMPI_Recon failed \n");
              }
           }
          
           gspeeds = (double*)malloc(
                              sizeof(double)
                              *
                              num_of_computers
           );

           if (gspeeds == NULL)
           {
              printf("(%d): Cannot allocate gspeeds\n", gme);
           }

           rc = HMPI_Group_performances(
                    HMPI_COMM_WORLD_GROUP,
                    gspeeds
           );
    
           if (rc != HMPI_OK)
           {
              printf(
                 "MAIN: Problems getting the performances"
                 "...Exiting...\n"
              );
           }

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              w = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
              );
   
              if (w == NULL)
              {
                 printf("(%d): Cannot allocate w\n", gme);
              }
   
              h = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q*p*q)
              );

              if (h == NULL)
              {
                 printf("(%d): Cannot allocate h\n", gme);
              }

              trow = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (trow == NULL)
              {
                 printf("(%d): Cannot allocate trow\n", gme);
              }

              tcol = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (tcol == NULL)
              {
                 printf("(%d): Cannot allocate tcol\n", gme);
              }

              rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    gspeeds,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
              );
   
              if (rc != HMPI_OK)
              {
                 printf("(%d): Problems partitioning\n", gme);
              }

              if (VERBOSE > 0)
              {
                 Print_rectangle_2d_local_stdin(
                  p, q, MSIZE, MSIZE,
                  w, h
                 );
              }

              int *model_params;
              int model_count;

              model_count = 1+1+1+1+(p*q*p*q)+(p*q*p*q);
              model_params = (int*)malloc(
                                   sizeof(int)
                                   *
                                   model_count
              );
    
              if (model_params == NULL)
              {
                 printf("No memory to allocate performanc model parameters\n");
              }          

              model_params[0]                      = p; 
              model_params[1]                      = q; 
              model_params[1+1]                    = n; 
              model_params[1+1+1]                  = kb; 

              for (i = 0; i < (p*q*p*q); i++) 
              {         
                 model_params[1+1+1+1+i]           = w[i];
                 model_params[1+1+1+1+(p*q*p*q)+i] = h[i];
              } 

              double timeof = HMPI_Timeof(
                                  &MPC_NetType_hetpdgemm,
                                  model_params,
                                  model_count
              );

              free(model_params);
              free(w);
              free(h);
              free(trow);
              free(tcol);

              printf(
                 "TIMEOF: Process arrangement (%d,%d) block size %d time(sec) %f\n", 
                 p, q,
                 kb,
                 timeof
              );
           }

           free(gspeeds);
       }

       HMPI_Finalize(0);
    }
