
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "scalapack_pzgemm_parameters.h"

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCB[DLEN_], DESCC[DLEN_], rc;

    char TRANSAC = TRANSAI;
    char *TRANSA = &TRANSAC;
    char TRANSBC = TRANSBI;
    char *TRANSB = &TRANSBC;

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    {
       double time, min_time = 1.7976931348623157E+308;
       int OPTIMALDATAB, OPTIMALDATAR;
       int OPTIMALDATABMAX = 1024;
       int OPTIMALALGOB    = 32;

       for (OPTIMALDATAB = OPTIMALALGOB; OPTIMALDATAB <= OPTIMALDATABMAX; OPTIMALDATAB+=32)
       {
           if (TRANSAC == 'N')
           {
              DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
              DESCA[M_]     = M;
              DESCA[N_]     = K;
              DESCA[IMB_]   = OPTIMALDATAB;
              DESCA[INB_]   = OPTIMALDATAB;
              DESCA[MB_]    = OPTIMALDATAB;
              DESCA[NB_]    = OPTIMALDATAB;
              DESCA[RSRC_]  = 0;
              DESCA[CSRC_]  = 0;
           }
           else
           {
              DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
              DESCA[M_]     = K;
              DESCA[N_]     = M;
              DESCA[IMB_]   = OPTIMALDATAB;
              DESCA[INB_]   = OPTIMALDATAB;
              DESCA[MB_]    = OPTIMALDATAB;
              DESCA[NB_]    = OPTIMALDATAB;
              DESCA[RSRC_]  = 0;
              DESCA[CSRC_]  = 0;
           }

           if (TRANSBC == 'N')
           {
              DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
              DESCB[M_]     = K;
              DESCB[N_]     = N;
              DESCB[IMB_]   = OPTIMALDATAB;
              DESCB[INB_]   = OPTIMALDATAB;
              DESCB[MB_]    = OPTIMALDATAB;
              DESCB[NB_]    = OPTIMALDATAB;
              DESCB[RSRC_]  = 0;
              DESCB[CSRC_]  = 0;
           }
           else
           {
              DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
              DESCB[M_]     = N;
              DESCB[N_]     = K;
              DESCB[IMB_]   = OPTIMALDATAB;
              DESCB[INB_]   = OPTIMALDATAB;
              DESCB[MB_]    = OPTIMALDATAB;
              DESCB[NB_]    = OPTIMALDATAB;
              DESCB[RSRC_]  = 0;
              DESCB[CSRC_]  = 0;
           }

           DESCC[DTYPE_] = BLOCK_CYCLIC_2D_INB;
           DESCC[M_]     = M;
           DESCC[N_]     = N;
           DESCC[IMB_]   = OPTIMALDATAB;
           DESCC[INB_]   = OPTIMALDATAB;
           DESCC[MB_]    = OPTIMALDATAB;
           DESCC[NB_]    = OPTIMALDATAB;
           DESCC[RSRC_]  = 0;
           DESCC[CSRC_]  = 0;

           //hscal_set_debug(HSCAL_LOG_VERBOSE);

           rc = hscal_pzgemm_recon(
                     TRANSA,
                     TRANSB,
                     &M,
                     &N,
                     &K,
                     (double*)&ALPHA,
                     &IA,
                     &JA,
                     DESCA,
                     &IB,
                     &JB,
                     DESCB,
                     (double*)&BETA,
                     &IC,
                     &JC,
                     DESCC
           );

           if (rc != HMPI_SUCCESS)
           {
              return rc;
           }

           if (HMPI_Is_host())
           {
              int *model_params, model_count;

              rc = hscal_pack_pzgemm_modelp_noauto_optimal_b(
                        TRANSA,
                        TRANSB,
                        &M,
                        &N,
                        &K,
                        (double*)&ALPHA,
                        &IA,
                        &JA,
                        DESCA,
                        &IB,
                        &JB,
                        DESCB,
                        (double*)&BETA,
                        &IC,
                        &JC,
                        DESCC,
                        &OPTIMALALGOB,
                        &nprow,
                        &npcol,
                        &model_params,
                        &model_count
              );

              if (rc != HSCAL_SUCCESS)
              {
                 return rc;
              }

              time = HMPI_Timeof(
                         hscal_model_pzgemm_optimal_b,
                         model_params,
                         model_count
              );

              printf("OPTIMALDATAB=%d, Time=%f\n", OPTIMALDATAB, time);

              if (time < min_time)
              {
                 min_time = time;
                 OPTIMALDATAR = OPTIMALDATAB;
              }

              free(model_params);
           }

           HMPI_Barrier(HMPI_COMM_WORLD_GROUP);
        }

        if (HMPI_Is_host())
        {
           printf("Optimal data block size is %d\n", OPTIMALDATAR);
        }
    }

    HMPI_Finalize(0);
}
